function [P, f] = ELSA(objective_fct, threshold, N, lb, ub, stopeval, varargin)
% [P, f] = ELSA(objective_fct, threshold, N, lb, ub, stopeval, varargin)
%
% Runs the ELSA algorithm for the specified objective function with the specified
% level set threshold. N is the number of dimensions, lb and ub are the lower and
% upper bound of the search space. stopeval is the evaluation budget.
%
% Authors: Johannes W. Kruisselbrink, Michael T.M. Emmerich, Andre H. Deutz

	% Parse the inputs
	p = inputParser;
	p.addRequired('objective_fct');
	p.addRequired('threshold');
	p.addRequired('N');
	p.addRequired('lb');
	p.addRequired('ub');
	p.addRequired('stopeval');
	p.addOptional('visualization', false);
	p.addOptional('diversity_indicator', @(P, f, threshold) D_ADI_augmented(P, f, N, lb, ub, threshold));
	p.addOptional('bch_fct', @reflect_bc);
	p.parse(objective_fct, threshold, N, lb, ub, stopeval, varargin{:});

	% Set the algorithm parameters
	objective_fct = p.Results.objective_fct;
	threshold = p.Results.threshold;
	N = p.Results.N;
	lb = p.Results.lb;
	ub = p.Results.ub;
	stopeval = p.Results.stopeval;
	visualization = p.Results.visualization;
	diversity_indicator = p.Results.diversity_indicator;
	bch_fct = p.Results.bch_fct;

	% Algorithm parameter
	mu = 20;
	sigma = mean(ub - lb) / sqrt(N);
	sigma = mean(ub - lb) / 30;

	% Counters
	evalcount = 0;
	gencount = 0;

	% Generate initial population
	for k = 1:mu
		xp(:,k) = lb + (ub - lb) .* rand(N,1);
		xp(:,k) = [-0.84, -2.04]' + 0.0000001 * rand(N,1);
		fp(k) = feval(objective_fct, xp(:,k)');
		evalcount = evalcount + 1;
	end

	% Optimization loop
	while (evalcount < stopeval)

		% Increase generation counter
		gencount = gencount + 1;

		% Generate offspring
		if (rand() < 0)
			% Random generation
			xo = lb + (ub - lb) .* rand(N,1);
		else
			% Generate from parent
			p = ceil(mu * rand());
			xo = xp(:,p) + sigma * randn(N,1);
		end
		xo = feval(bch_fct, xo, lb, ub);
		fo = feval(objective_fct, xo');
		evalcount = evalcount + 1;

		% Combine the parents and offspring in one set
		P = [xp,xo];
		f = [fp,fo];

		% Compute the diversity indicator contributions
		% Remove the one that has the biggest ADI contribution
		DI_contributions = compute_DI_contributions(P, f, threshold, diversity_indicator);

		% Sort based on contributions
		[sortDIC, sortindex] = sort(DI_contributions, 'ascend');

		% Eliminate the solution with the least contribution
		xp = P(:,sortindex(1:mu));
		fp = f(sortindex(1:mu));

		% Compute the diversity of the current generation
		generation_diversity(gencount) = feval(diversity_indicator, xp, fp, threshold);

		% Plot the statistics of the current generation
		if (visualization)
			subplot(1,2,1)
			plot(generation_diversity(1:gencount));
			subplot(1,2,2)
			plot_level_set_approximation_landscape(objective_fct, threshold, N, lb, ub, P);
			drawnow
		end
	end

	P = xp;
	f = fp;

end

